#include <mpi.h>
#include <omp.h>
#include <functional>
#include <algorithm>
#include <numeric>
#include <unistd.h>
#include "defs.h"

extern void peer2peer_exchange(const std::vector<int> &params, olocal &output);
extern void hip_memcopy(const std::vector<int> &params, olocal &output);
extern int hip_stream(const std::vector<int> &params, olocal &output);
extern void core_affinity(const std::vector<int> &params, olocal &output);
extern void nic_affinity(const std::vector<int> &params, olocal &output);
extern void device_affinity(const std::vector<int> &params, olocal &output);

enum check_t { MAX, MIN, AVG, NUM };

bool do_check(int rank, const std::vector<oglobal> &output, const std::string &name, check_t type, double value, size_t num = 0) {
    bool result = false;
    for (const auto &gl : output) {
        if (gl.header == name) {
            if (gl.values.size() == 0) {
                print(rank, std::string("CHECK: No values for: ") + name);
                return false;;
            }
            double control_value = 0;
			std::string stype;
            switch (type) {
                case check_t::MAX: stype = "MAX"; control_value = *std::max_element(gl.values.begin(), gl.values.end()); result = true; break;
                case check_t::MIN: stype = "MIN"; control_value = *std::max_element(gl.values.begin(), gl.values.end()); result = true; break;
                case check_t::AVG: stype = "AVG"; control_value = std::reduce(gl.values.begin(), gl.values.end()) / (double)gl.values.size(); result = true; break;
                case check_t::NUM: stype = "NUM";
							  if (gl.values.size() < num) {
                                  print(rank, std::string("CHECK: Index out of bounds for: ") + name);  
                                  break;
                              }
                              control_value = gl.values[num]; 
                              result = true;
                              break;
            }
            if (result && control_value < value) {
				print(rank, std::string("CHECK: Failed for: ") + name + " type: " + stype + " expected: " + std::to_string(value) + " obtained: " + std::to_string(control_value));
                result = false;
			}
            return result;
        }    
    }
    return result;
}

void do_test(std::function<void(const std::vector<int> &, olocal &output)> fn, const std::vector<int> &params, olocal &output) {
#pragma omp parallel 
    {    
        for (int j = 0; j < 1; j++) {
#pragma omp master
            {
                fn(params, output);
            }
        }
    }
    MPI_Barrier(MPI_COMM_WORLD);
}

void print_final_output(int rank, const std::vector<oglobal> &output) {
    for (const auto &og : output) {
        std::stringstream ss_outer;
        ss_outer << og.header << ": {";
        std::stringstream ss_inner;
        ss_inner << std::setprecision(1) << std::fixed;
        for (const auto value : og.values) {
            ss_inner << "," << value;
        }
        ss_outer << ss_inner.str().erase(0, 1) << "}";
        print(rank, ss_outer.str());
    }
}

void local2output(olocal &output_local, std::vector<oglobal> &output) {
    if (output_local.values.empty())
        return;
    std::vector<double> aggregator;
    for (const auto &v : output_local.values) {
        for (const auto value : v) {        
            aggregator.push_back(value);
        }
    }
    output.push_back({output_local.header, aggregator});
    output_local.values.resize(0);
}

int main(int argc, char **argv)
{
    std::vector<oglobal> output;
    MPI_Init(&argc, &argv);
    alarm(170); // NOTE: this alarm is to interrupt test in case of deadlock
    int size, rank;
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    int n = omp_get_max_threads();
    std::cout << ">> rank=" << rank << " omp nthreads=" << n << std::endl;
    MPI_Barrier(MPI_COMM_WORLD);
    olocal output_local;
    do_test(core_affinity, {rank}, output_local);
    do_test(device_affinity, {rank}, output_local);
    do_test(nic_affinity, {rank}, output_local);
    do_test(hip_stream, {rank}, output_local);
    local2output(output_local, output);
    const int ngpus = 8;
    for (const int use_pinned_memory : {1,0}) { 
        for (int distance = 1; distance < ngpus * 2; distance *= 2) {
            do_test(hip_memcopy, {rank, distance, ngpus, use_pinned_memory}, output_local);
        }
        local2output(output_local, output);
    }
    for (const int use_gpu_memory : {1,0}) { // {1,0} for both GPU and CPU exch tests
        std::string kind = (use_gpu_memory ? "Device2Device" : "Host2Host");
		print(rank, kind + " MPI exchange, intra-node:");
		for (int distance = 1; distance < ngpus; distance *= 2) {
			do_test(peer2peer_exchange, {rank, size, distance, ngpus, use_gpu_memory}, output_local);
		}
        local2output(output_local, output);
        if (size > ngpus) {
            print(rank, kind + "MPI exchange, cross-node:");
            for (int distance = 0; distance < ngpus; distance++) {
                do_test(peer2peer_exchange, {rank, size, distance + ngpus, ngpus, use_gpu_memory}, output_local);
            }
            local2output(output_local, output);
        }
    }
    print_final_output(rank, output);
/*
rank=0: GPU_STREAM: {1279.1,1280.7,1245.8,1245.8}
rank=0: H2D_D2H_pinned: {22.9,23.6,23.5,23.7,23.5,23.7,23.5,23.7}
rank=0: H2D_D2H_pageable: {6.1,5.8,6.0,5.9,6.1,6.1,6.1,6.2}
rank=0: MPI_P2P_intranode_GPU: {135.1,35.0,34.9}
rank=0: MPI_P2P_crossnode_GPU: {21.3,13.6,21.3,14.3,21.3,13.6,21.2,22.3}
rank=0: MPI_P2P_intranode_CPU: {18.9,18.7,18.0}
rank=0: MPI_P2P_crossnode_CPU: {18.0,14.9,18.2,13.9,17.9,14.2,18.2,22.3}
*/
	bool result = true;
	result = result & do_check(rank, output, "GPU_STREAM", AVG, 1200.0);
	result = result & do_check(rank, output, "GPU_STREAM", MIN, 1200.0);
	result = result & do_check(rank, output, "H2D_D2H_pinned", AVG, 18.0);
	result = result & do_check(rank, output, "H2D_D2H_pinned", MIN, 17.0);
	result = result & do_check(rank, output, "H2D_D2H_pageable", AVG, 3.0);
	result = result & do_check(rank, output, "H2D_D2H_pageable", MIN, 3.0);
	result = result & do_check(rank, output, "MPI_P2P_intranode_GPU", MAX, 130.0);
	result = result & do_check(rank, output, "MPI_P2P_intranode_GPU", MIN, 30.0);
	result = result & do_check(rank, output, "MPI_P2P_crossnode_GPU", MAX, 20.0);
	result = result & do_check(rank, output, "MPI_P2P_crossnode_GPU", MIN, 12.0);
	result = result & do_check(rank, output, "MPI_P2P_intranode_CPU", AVG, 18.0);
	result = result & do_check(rank, output, "MPI_P2P_intranode_CPU", MIN, 17.0);
	result = result & do_check(rank, output, "MPI_P2P_crossnode_CPU", MAX, 17.0);
	result = result & do_check(rank, output, "MPI_P2P_crossnode_CPU", MIN, 12.0);
    if (!result) {
        print(rank, "CHECK FAILED");
    }
    int mpiresult = (result?1:0);
    MPI_Allreduce(MPI_IN_PLACE, &mpiresult, 1, MPI_INT, MPI_LAND, MPI_COMM_WORLD);
    if (!mpiresult) {
        print(rank, "CHECK FAILED ON ONE OR RANKS");
    }
    MPI_Finalize();
    return !mpiresult;
}
